# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2019 uib GmbH <info@uib.de>

"""
Commandline interface.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""

import argparse
import json
from itertools import chain

from . import __version__
from .config import convertConfig, readConfig
from .connector import convertComputers
from .grouphandling import DNbasedGroupHandler, UCSatSchoolGroupHandler
from .ldap import (
    createSearchQuery, convertElementToDict, checkConnectionToDirectory,
    directoryConnection, readDirectory)
from .logger import LOGGER, configureLogger
from .opsi import (
    modulesCheck, checkConnectionToOpsi, checkRootGroup, createJSONRPC,
    getIdsOfHostsInOpsi, getSession, updateOpsi)


__all__ = ('runConnector', 'parseCommandlineArguments')


def runConnector():
    cliArguments = parseCommandlineArguments()
    configureLogger(LOGGER, level=cliArguments.logLevel * 10)
    config = readConfig(cliArguments.config)
    config = convertConfig(config)
    untouchedConfig = config.copy()

    rootGroup = config['behaviour']['opsiRootDir']
    LOGGER.debug("root group: {}".format(rootGroup))

    if not rootGroup:
        raise ValueError("Invalid configuration for behaviour -> root_dir_in_opsi")

    # Remove the value before connecting to avoid handing an unexpected
    # parameter to the connection function.
    exitOnError = config['opsi'].pop('exitOnError', False)

    if cliArguments.checkOpsi:
        checkConnectionToOpsi(config['opsi'], rootGroup)
        return  # No further processing

    searchQuery = createSearchQuery(
        computerQuery=config['directory'].pop('searchQueryForComputers', None)
    )
    directoryConfig = config['directory']
    pagedSearch = directoryConfig['connectionOptions'].pop('paged_search_limit', 0)
    dirConSettings = {
        "address": directoryConfig['address'],
        "user": directoryConfig['user'],
        "password": directoryConfig['password'],
        "connectionOptions": directoryConfig['connectionOptions']
    }
    with directoryConnection(**dirConSettings) as dirConn:
        elements = readDirectory(
            dirConn,
            searchBase=directoryConfig['searchBase'],
            searchQuery=searchQuery,
            mapping=config['mapping'],
            identifyingAttribute=directoryConfig['identifyingAttribute'],
            pagedSearch=pagedSearch
        )

        elements = (convertElementToDict(element) for element in elements)
        if cliArguments.checkDirectory:
            checkConnectionToDirectory(elements)
            return  # No further processing

        writeChanges = config['behaviour'].get('writeChangeToOpsi', False)

        with getSession(**config['opsi']) as opsiSession:
            modulesCheck(opsiSession)
            checkRootGroup(opsiSession, rootGroup)
            hostIDs = getIdsOfHostsInOpsi(opsiSession)

            groupHandlingStrategy = config['behaviour'].pop('groupHandling', 'dn')
            LOGGER.debug("Group handling strategy: {!r}".format(groupHandlingStrategy))
            if groupHandlingStrategy.strip().lower() == 'ucsatschool':
                LOGGER.info("Configured for handling groups in UCS@school-style.")
                groupHandler = UCSatSchoolGroupHandler(rootGroup, opsiSession, dirConn, untouchedConfig)
            else:
                groupHandler = DNbasedGroupHandler(rootGroup, opsiSession)

            clientActions = convertComputers(
                elements, hostIDs, groupHandler, opsiSession,
                config['mapping'], **config['behaviour']
            )
            postComputerActions = groupHandler.postClientProcessing()

            actions = chain(clientActions, postComputerActions)
            updateOpsi(opsiSession, actions, writeChanges=writeChanges, exitOnError=exitOnError)

            opsiSession.get(params=json.dumps(createJSONRPC(method="backend_exit")))


def parseCommandlineArguments(args=None):
    parser = argparse.ArgumentParser(description="opsi directory connector")
    parser.add_argument('--version', action="version", version=__version__)
    parser.add_argument('--log-level', '-l', dest='logLevel',
                        help=(
                            'Sets how much information will be logged. '
                            'From 1 (debug) to 5 (critical)'
                        ),
                        type=int, choices=[1, 2, 3, 4, 5], default=3)
    checks = parser.add_mutually_exclusive_group()
    checks.add_argument('--check-directory', dest="checkDirectory",
                        action="store_true",
                        help=(
                            'Check if a connection to the directory '
                            'can be established and if items will be '
                            'received.')
                        )
    checks.add_argument('--check-opsi', dest="checkOpsi",
                        action="store_true",
                        help=(
                            'Check if a connection to the opsi server '
                            'can be established.')
                        )
    parser.add_argument('--config', help='Path to the config.', required=True)
    return parser.parse_args(args=args)


if __name__ == '__main__':
    runConnector()
