# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2018-2019 uib GmbH <info@uib.de>
"""
Configuration handling.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""

import json

__all__ = ('readConfig', 'convertConfig')


def readConfig(path):
    with open(path, 'r') as configFile:
        return json.load(configFile)


def convertConfig(config):
    """
    Convert the given `config` into a format that allows passing the \
config to our processing functions.
    """
    mappings = {
        "directory": {
            "search_base": "searchBase",
            "search_query_computers": "searchQueryForComputers",
            "connection_options": "connectionOptions",
            "identifying_attribute": "identifyingAttribute",
            "group_strategy": "groupStrategy",
        },
        "behaviour": {
            "group_handling": "groupHandling",
            "update_existing_clients": "updateExistingClients",
            "prefer_location_from_directory": "preferLocationFromDirectory",
            "root_dir_in_opsi": "opsiRootDir",
            "write_changes_to_opsi": "writeChangeToOpsi",
        },
        "opsi": {
            "connection_options": "connectionOptions",
            "exit_on_error": "exitOnError",
        }
    }

    newConfig = {}

    for branch, configs in config.items():
        newConfig[branch] = {}

        for key, value in configs.items():
            try:
                newKey = mappings[branch][key]
                newConfig[branch][newKey] = value
            except KeyError:
                newConfig[branch][key] = value

    removePrefixFromMapping(newConfig, "client", "c")

    for branch in ('directory', 'opsi'):
        try:
            path = newConfig[branch].pop('passwordFile')
            if path:
                with open(path, 'r') as credentials:
                    newConfig[branch]['password'] = credentials.read()
        except KeyError:
            pass

    if 'identifyingAttribute' not in newConfig['directory']:
        newConfig['directory']['identifyingAttribute'] = 'dn'

    try:
        newConfig['mapping']['groupName'] = newConfig['mapping'].pop('group_name')
    except KeyError:
        newConfig['mapping']['groupName'] = {}

    for unwantedKey in ("search_query_groups", "searchQueryForGroups"):
        try:
            del newConfig["directory"][unwantedKey]
        except KeyError:
            pass  # Missing is not a problem

    try:
        del newConfig['mapping']['group']
    except KeyError:
        pass

    if 'behaviour' not in newConfig:
        newConfig['behaviour'] = {}

    try:
        newConfig['behaviour']['groupHandling'] = newConfig['directory'].pop('groupStrategy', newConfig['behaviour']['groupHandling']).lower()
    except KeyError:
        newConfig['behaviour']['groupHandling'] = 'dn'

    return newConfig


def removePrefixFromMapping(config, objectType, prefix):
    "Removes the prefix `prefix` from the mapping configued in `config` for `objectType`."
    mapping = config['mapping'][objectType]
    for key in list(mapping.keys()):
        if key.startswith(prefix):
            modifiedKey = key[1:]
            mapping[modifiedKey] = mapping[key]
            del mapping[key]

    config['mapping'][objectType] = mapping
