# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2019 uib GmbH <info@uib.de>

"""
Connector logic.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""

from collections import namedtuple

from .opsi import createJSONRPC, getHostFromOpsi
from .logger import LOGGER
from .utilities import areValuesDifferent, isFqdn

try:
    from ldap3 import LDAPKeyError
except ImportError:
    # Probably ldap3 > 2.0
    from ldap3.core.exceptions import LDAPKeyError

ClientMapping = namedtuple("ClientMapping", "id description notes hardwareAddress ipAddress inventoryNumber oneTimePassword")


def convertComputers(elements, hostsInOpsi, groupHandler,
                    session, attributeMapping,
                    updateExistingClients=False,
                    preferLocationFromDirectory=False,
                    identifyingAttribute='dn', **_):

    clientAttributes = createClientMapping(attributeMapping['client'])
    groupNameReplacements = attributeMapping['groupName']

    opsiHosts = set(hostsInOpsi)
    for element in elements:
        LOGGER.debug("Next element: {0!r}".format(element))

        try:
            dn = getDNFromElement(element, identifyingAttribute)
            parentGroups = getGroupsFromDN(element, dn, groupNameReplacements)

            clientId = getFQDN(element, clientAttributes.id, dn)

            if clientId in opsiHosts:
                LOGGER.info("Host {0!r} exists in opsi".format(clientId))
                if updateExistingClients:
                    LOGGER.info("Updating existing host {0!r}".format(clientId))
                    if preferLocationFromDirectory:
                        LOGGER.debug("Removing existing group assignments of {0!r}".format(clientId))
                        yield groupHandler.deleteGroupAssignments(clientId)

                    for groupAction in groupHandler.processGroups(clientId, parentGroups):
                        yield groupAction

                    yield getClientUpdateAction(session, clientId, element, clientAttributes._asdict())
                else:
                    LOGGER.debug("Skipped existing host {!r}".format(element))
            else:
                LOGGER.info("Host {0!r} does not exist in opsi.".format(clientId))
                yield getClientCreationAction(clientId, element, clientAttributes)
                opsiHosts.add(clientId)  # client is now in opsi

                for groupAction in groupHandler.processGroups(clientId, parentGroups):
                    yield groupAction
        except LDAPKeyError as error:
            LOGGER.warning("Processing {} failed: {}".format(element, error))
            LOGGER.debug("Full exception: {!r}".format(error))


def getDNFromElement(element, identifyingAttribute='dn'):
    try:
        return element[identifyingAttribute]
    except KeyError as kerr:
        LOGGER.debug("Failed to get identifying name from {0!r} (via attribute {attr!r}): {error!r}".format(element, error=kerr, attr=identifyingAttribute))
        LOGGER.critical("Unable to get the identifying name for the current element!")
        raise kerr


def getGroupsFromDN(element, dn, groupNameReplacements):
    """
    Read the groups from the DN of the client.

    :param element: The element in question
    :type element: dict
    :param dn: The DN of the element
    :type dn: str
    :param groupNameReplacements: The read group names might be replaced by this.
    :type groupNameReplacements: dict
    :return: groups in the dn with the groupnames being lowercase.
    :rtype: [str, ]
    """
    try:
        category = getAttributeFromElement(element, 'objectCategory', flatten=True)
        fullDN = "{0},{1}".format(category, dn)
    except KeyError:
        fullDN = dn

    groups = getGrouplist(fullDN)
    return replaceGroups(groups, groupNameReplacements)


def getGrouplist(dn):
    """
    Gets the groups from a DN.

    As a group every Organisational Unit (OU) qualifies.

    :type dn: str
    :returntype: [str, ]
    :return: groups in the dn with the groupnames being lowercase.
    """
    groups = []
    for element in dn.lower().split(","):
        if element.startswith("ou="):
            group = element.replace("ou=", "", 1).strip()
            groups.append(group)

    LOGGER.debug("Groups {0!r} extracted from {1}".format(groups, dn))
    return groups


def replaceGroups(groups, replacements):
    """
    Apply the `replacements` to `groups`.

    :type groups: list
    :param replacements: Replace every occurance of `key` with `value`.
    :type replacements: {str: str}
    :rtype: list
    """
    for index, group in enumerate(groups):
        try:
            replacement = replacements[group]
            groups[index] = replacement.lower().strip()
            LOGGER.debug("Replaced group {!r} with {!r}".format(group, replacement))
        except KeyError:
            pass

    return groups


def getFQDN(element, fqdnAttribute, dn):
    """
    Create FQDN for the element.

    In case `fqdnAttribute` does not supply a complete FQDN we
    make use of the DN obtained from `dn`.

    :param element: The element for which to create a FQDN.
    :type element: dict
    :param fqdnAttribute: The attribute where to look for the FQDN
    :type fqdnAttribute: str
    :param dn: The DN of the element.
    :type dn: str
    :return: The FQDN of the element.
    :rtype: str
    :raise KeyError: In case the attribute of the FQDN could not be read.
    """
    try:
        name = getAttributeFromElement(element, fqdnAttribute, flatten=True)
    except KeyError as kerr:
        LOGGER.debug("{0!r} during reading of the attribute for the client id.".format(kerr))
        LOGGER.critical("Unable to read the attribute {attr!r} specified for the client ID. Aborting.".format(attr=fqdnAttribute))
        raise kerr

    if not isFqdn(name):
        LOGGER.debug("Need to create a fqdn...")
        fqdn = "{0}.{1}".format(name, getDomain(dn))
    else:
        fqdn = name

    fqdn = fqdn.lower()  # opsi expected FQDN in lowercase.
    LOGGER.debug("Using {0!r} as fqdn".format(fqdn))

    return fqdn


def getAttributeFromElement(element, attribute, flatten=True):
    """
    Return the value for `attribute` in element.

    This is designed to work with different versions of dicts created
    from different versions of ldap3.

    :param element: The element to search the attribute for.
    :type element: dict
    :param attribute: The attribute to search for.
    :type attribute: str
    :param flatten: If the attributes holds many values this selects the first value to be returned.
    :type flatten: bool
    :raises KeyError: If no matching attribute is found.
    """
    attr = element[attribute]

    if flatten and isinstance(attr, list):
        return attr[0]

    return attr


def getDomain(dn):
    """
    Get the domain from the given dn.

    :type dn: str
    :return: the domain from the DN in format `domain.tld`
    :returntype: str
    """
    domainComponents = ("DC=", "dc=")
    domainParts = []
    for element in dn.split(","):
        for domainComponent in domainComponents:
            if element.startswith(domainComponent):
                domainParts.append(element.replace(domainComponent, ""))

    domain = ".".join(domainParts)
    LOGGER.debug("domain {0!r} created from {1}".format(domain, dn))
    return domain


def createClientMapping(clientAttributeMapping):
    return ClientMapping(
        clientAttributeMapping["id"],
        clientAttributeMapping.get("description", ""),
        clientAttributeMapping.get("notes", ""),
        clientAttributeMapping.get("hardwareAddress", ""),
        clientAttributeMapping.get("ipAddress", ""),
        clientAttributeMapping.get("inventoryNumber", ""),
        clientAttributeMapping.get("oneTimePassword", "")
    )


def getClientUpdateAction(session, clientId, element, clientAttributeMapping):
    opsiHost = getHostFromOpsi(session, clientId)

    updateNecessary = False
    for opsiAttribute, dirAttribute in clientAttributeMapping.items():
        if not dirAttribute or opsiAttribute == 'id':
            # We limit the updating to the attributes that have a
            # mapping and we also do not want override the id because
            # changing the ID would result in an different opsi object.
            continue

        valueInOpsi = opsiHost[opsiAttribute]
        valueInDirectory = convertToStringIfNotNone(element, clientAttributeMapping[opsiAttribute])

        LOGGER.debug("Comparing {0!r} (opsi) and {1!r} (directory)".format(valueInOpsi, valueInDirectory))
        if areValuesDifferent(valueInOpsi, valueInDirectory):
            LOGGER.debug("Difference detected. Updating {attr} of our opsi host.".format(attr=opsiAttribute))
            opsiHost[opsiAttribute] = valueInDirectory
            updateNecessary = True

    if updateNecessary:
        LOGGER.info("Updating attributes of host {0!r}".format(clientId))
        return createJSONRPC('host_insertObject', [opsiHost])


def convertToStringIfNotNone(element, attribute):
    """
    Read the value for `attribute` from `element`.

    If the value is not `None`, then it will be explicitely converted to str.
    If there is no attribute found `None` will be returned.
    If no attribute is given `None` will be returned.

    :type attribute: str
    :param attribute: The attribute to search for.
    :returntype: str or None
    """
    if not attribute:
        return None

    try:
        attr = getAttributeFromElement(element, attribute, flatten=True)
        if attr is not None:
            return str(attr)
    except IndexError as ierr:
        LOGGER.debug("No {attr!r} in {elem!r} probably not filled. Received {err!r}".format(attr=attribute, elem=element, err=ierr))
    except KeyError as kerr:
        LOGGER.debug("No {attr!r} in {elem!r}? Received {err!r}".format(attr=attribute, elem=element, err=kerr))

    return None


def getClientCreationAction(fqdn, element, clientAttributes):
    params = [
        fqdn,
        None,  # HostKey - not setting one
        convertToStringIfNotNone(element, clientAttributes.description),
        convertToStringIfNotNone(element, clientAttributes.notes),
        convertToStringIfNotNone(element, clientAttributes.hardwareAddress),
        convertToStringIfNotNone(element, clientAttributes.ipAddress),
        convertToStringIfNotNone(element, clientAttributes.inventoryNumber),
        convertToStringIfNotNone(element, clientAttributes.oneTimePassword),
        None,  # created
        '0000-00-00 00:00:00',  # lastSeen
    ]
    return createJSONRPC(method='host_createOpsiClient', params=params)
