# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2019 uib GmbH <info@uib.de>
"""
Handling of groups.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""
from .connector import getAttributeFromElement, getDomain
from .ldap import convertElementToDict, searchDirectory
from .logger import LOGGER
from .opsi import createJSONRPC, getGroupsFromOpsi, getClientGroupMapping
from .utilities import listReplace

__all__ = ('DNbasedGroupHandler', 'UCSatSchoolGroupHandler')

NAME_ADDITION = '_'


class BaseGroupHandler:

    def __init__(self, rootGroup, opsiSession):
        """
        Create a new handler.

        :param rootGroup: The group to use as root.
        :type rootGroup: str
        :param opsiSession: A session connected to the opsi API.
        """
        self.rootGroup = rootGroup

        self._session = opsiSession
        self._clientGroupMapping = getClientGroupMapping(self._session)
        self._groupsInOpsi = getGroupsFromOpsi(self._session)

    def deleteGroupAssignments(self, clientId):
        """
        Remove group assignments of the given client.

        :param clientId: The ID of the client whose assignments should be removed.
        :type clientId: str
        """
        def isFromRootDir(groupname):
            if groupname is None:
                return False
            elif groupname == self.rootGroup:
                return True

            for grp in self._groupsInOpsi:
                if groupname == grp["id"]:
                    result = isFromRootDir(grp["parentGroupId"])
                    LOGGER.debug("isFromRootDir parentGroupId {0!r}: {1}".format(grp["parentGroupId"], result))
                    return result

            return False

        for clientGroupMapping in self._clientGroupMapping:
            if clientId == clientGroupMapping["objectId"]:
                groupname = clientGroupMapping["groupId"]
                LOGGER.debug("Found group {1!r} for client {0!r}".format(clientId, groupname))
                if isFromRootDir(groupname):
                    return createJSONRPC(
                        method="objectToGroup_deleteObjects",
                        params=[clientGroupMapping]
                    )

    def processGroups(self, clientId, parentGroups):
        raise NotImplementedError("Has to be implemented by subclass")


class DNbasedGroupHandler(BaseGroupHandler):
    """
    Handling groups with the DN of an client as the base.

    Groups are created based on the OUs a client is in.
    """

    def processGroups(self, clientId, parentGroups):
        LOGGER.info("Processing groups of {!r}".format(clientId))
        groupsChanged = False
        for groupAction in processGroupsOfHost(clientId, parentGroups, self.rootGroup, self._groupsInOpsi):
            yield groupAction
            groupsChanged = True

        if groupsChanged:
            self._groupsInOpsi = getGroupsFromOpsi(self._session)

    def postClientProcessing(self):
        yield  # Doing nothing, but providing an iterator


def processGroupsOfHost(clientId, parentGroups, rootGroup, groupsInOpsi):
    try:
        if not parentGroups:
            raise IndexError("No parents for {}.".format(clientId))

        for action in processParentList(parentGroups, rootGroup, groupsInOpsi):
            yield action

        parent = parentGroups[0]
        LOGGER.debug("Parent for {!r} is {!r}".format(clientId, parent))
        yield createHostGroupAssignment(parent, clientId)
    except IndexError:
        LOGGER.debug("No groups for {0!r} found.".format(clientId))


def processParentList(parentList, rootGroup, groupsInOpsi):
    """
    Make sure that the groups in `parentList` exist in opsi.

    :param parentList: List of group parents - usually a _path_.
    :type parentList: [str, ]
    :param rootGroup: The root group under which new groups are to be created.
    :type rootGroup: str
    :param groupsInOpsi: Groups already existing in opsi.
    :type groupsInOpsi: [dict, ]
    """
    def createGroup(groupName):
        LOGGER.info("Creating missing group {0!r}".format(groupName))
        action = createHostGroupCommand(groupName, parentGroupId)
        groupsInOpsi.append(getGroupDefinition(action))
        return action

    reversedGroupList = parentList[::-1]

    for groupName in reversedGroupList:  # start at the group near the root.
        parentGroupId = getParent(groupName, parentList, rootGroup)
        LOGGER.debug("Found parent {0!r}".format(parentGroupId))

        if not groupExistsInOpsi(groupName, groupsInOpsi):
            yield createGroup(groupName)
        else:
            if hasGroupRightParent(groupName, parentGroupId, groupsInOpsi):
                LOGGER.debug("Group {0!r} exists in opsi and is at the correct location.".format(groupName))
                continue

            originalGroupName = groupName

            while groupExistsInOpsi(groupName, groupsInOpsi):
                if hasGroupRightParent(groupName, parentGroupId, groupsInOpsi):
                    break  # this is the correct group
                else:
                    groupName = groupName + NAME_ADDITION

            if groupName != originalGroupName:
                LOGGER.debug("Need to correct name of parent...")
                parentList = listReplace(parentList, originalGroupName, groupName)
                LOGGER.debug("Parents are now: {0!r}".format(parentList))

            if not groupExistsInOpsi(groupName, groupsInOpsi):
                yield createGroup(groupName)


def getParent(element, groups, rootGroup):
    """
    Searching for the parent of `element` in `groups`.
    Case is ignored during the search.

    If no parent is found then `rootGroup` will be returned.

    :type element: str
    :type groups: [str, ]
    :type rootGroup: str
    """
    LOGGER.debug('Looking for {0!r} in groups {1!r}'.format(element, groups))

    normalizedGroups = [group.lower() for group in groups]

    try:
        positionOfCurrent = normalizedGroups.index(element.lower())
        groupLen = len(groups)
        if positionOfCurrent <= (groupLen - 1):
            if (positionOfCurrent + 1) <= (groupLen - 1):
                parent = groups[positionOfCurrent + 1]
                LOGGER.debug("Returning parent {0!r}".format(parent))
                return parent
    except ValueError:
        LOGGER.debug("{!r} not in groups.".format(element))

    LOGGER.debug("Returning root {!r}".format(rootGroup))
    return rootGroup


def groupExistsInOpsi(name, existingGroups):
    """
    Check if a group with `name` does already exist in opsi.

    :param name: Name of the group to check for.
    :type name: str
    :param existingGroups: An iterable containing dicts with \
information about the existing groups.
    :type existingGroups: [dict, ]
    """
    lowecaseName = name.lower()
    exists = any(lowecaseName == group['id'] for group in existingGroups)
    LOGGER.debug("Exists group {0!r} in opsi? {1}".format(name, exists))
    return exists


def createHostGroupCommand(name, parentGroupId, description=None, notes=None, allowEmptyParent=False):
    if not name:
        raise ValueError("Missing group id!")

    if not allowEmptyParent and not parentGroupId:
        raise ValueError("Missing ID for parent or root group.")

    if name == parentGroupId:
        raise ValueError("Groupname is same as parent groupname. Forbidden!")

    return createJSONRPC(
        method="group_insertObject",
        params=[{
            "id": name,
            "description": description,
            "notes": notes,
            "parentGroupId": parentGroupId,
            "type": "HostGroup"
        }]
    )


def getGroupDefinition(jsonRpcCall):
    definition = jsonRpcCall['params'][0]
    definition['id'] = definition['id'].lower()  # opsi requirement
    return definition


def hasGroupRightParent(groupId, parentGroupId, existingGroups):
    """
    Checks if the group with id `groupId` and parent `parentGroupId` \
exist in `existingGroups`.

    :type groupId: str
    :type parentGroupId: str
    :type existingGroups: [dict, ]
    :returns: `True` if a match is found. `False` otherwise.
    :rtype: bool
    """
    LOGGER.debug("Checking if {0!r} has parent {1!r}".format(groupId, parentGroupId))
    LOGGER.debug("Existing groups: {0}".format(existingGroups))

    name = groupId.lower()
    for group in iterateHostGroups(existingGroups):
        if group['id'].lower() == name and group['parentGroupId'] is not None:
            parentMatches = group['parentGroupId'].lower() == parentGroupId.lower()
            LOGGER.debug("Parent matches? {0}".format(parentMatches))

            # Group ID is unique and we can stop searching now.
            return parentMatches

    LOGGER.debug("No match of group {0} with parent {1} found.".format(groupId, parentGroupId))
    return False


def iterateHostGroups(groups):
    """
    Iterate over all groups that are of type `HostGroup`.

    :param groups:
    :type groups: [dict, ]
    :rtype: dict
    """
    for group in groups:
        if group['type'] == 'HostGroup':
            yield group


def createHostGroupAssignment(groupname, clientId):
    return createJSONRPC(
        "objectToGroup_createObjects",
        [{
            "groupId": groupname,
            "objectId": clientId,
            "groupType": "HostGroup",
            "type": "ObjectToGroup"
        }]
    )


class UCSatSchoolGroupHandler(BaseGroupHandler):

    def __init__(self, rootGroup, opsiSession, directoryConnection, config):
        super().__init__(rootGroup, opsiSession)

        self._directoryConnection = directoryConnection
        self._config = config

        self._clientsForProcessing = set()

    def processGroups(self, clientId, parentGroups):
        LOGGER.debug("Saving {!r} for later group processing...".format(clientId))
        self._clientsForProcessing.add(clientId)
        yield

    def postClientProcessing(self):
        for action in self.processUCSAtSchoolGroups():
            yield action

    def processUCSAtSchoolGroups(self):
        LOGGER.debug("Starting to process schools...")
        searchBase = self._config['directory']['searchBase']
        schoolAttr = ['dn', 'ou', 'displayName']
        schoolQuery = '(objectClass=ucsschoolOrganizationalUnit)'
        pagedSearch = self._config['directory']['connectionOptions'].pop('paged_search_limit', 0)

        schools = searchDirectory(self._directoryConnection, searchBase, schoolQuery, schoolAttr, pagedSearch)
        schools = (convertElementToDict(school) for school in schools)
        for school in schools:
            baseGroupId = getAttributeFromElement(school, 'ou')

            try:
                description = getAttributeFromElement(school, 'displayName')
            except KeyError:
                description = None

            if not groupExistsInOpsi(baseGroupId, self._groupsInOpsi):
                LOGGER.info("Creating group for school {0!r}".format(baseGroupId))
                # Groups aren't part of the directory to allow having
                # a client in multiple rooms
                createGroupAction = createHostGroupCommand(
                    baseGroupId,
                    None,
                    allowEmptyParent=True,
                    description=description
                )
                self._groupsInOpsi.append(getGroupDefinition(createGroupAction))
                yield createGroupAction

            for action in self.processGroupsOfSchool(school):
                yield action

        LOGGER.debug("Done processing schools.")

    def processGroupsOfSchool(self, schoolElement):
        # For this to work we need to have the searchBase set to the root.
        schoolBase = getAttributeFromElement(schoolElement, 'dn')
        LOGGER.debug("Searching for groups of {!r}".format(schoolBase))
        searchBase = "cn=raeume,cn=groups,{school}".format(school=schoolBase)

        parentGroupId = getAttributeFromElement(schoolElement, 'ou')
        searchQuery = '(objectClass=univentionGroup)'
        attr = ['dn', 'cn', 'uniqueMember', 'description']

        pagedSearch = self._config['directory']['connectionOptions'].pop('paged_search_limit', 0)

        groups = searchDirectory(self._directoryConnection, searchBase, searchQuery, attr, pagedSearch)
        groups = (convertElementToDict(group) for group in groups)
        for group in groups:
            LOGGER.debug("Found group mapping {!r}".format(group))

            groupId = getAttributeFromElement(group, 'cn', flatten=True).lower()
            LOGGER.debug("Created groupId: {!r}".format(groupId))

            try:
                description = getAttributeFromElement(group, 'description')
            except KeyError:
                description = None

            if not groupExistsInOpsi(groupId, self._groupsInOpsi):
                LOGGER.info("Creating missing group {0!r}".format(groupId))
                createGroupAction = createHostGroupCommand(groupId, parentGroupId, description=description)
                self._groupsInOpsi.append(getGroupDefinition(createGroupAction))
                yield createGroupAction
            else:
                if not hasGroupRightParent(groupId, parentGroupId, self._groupsInOpsi):
                    while groupExistsInOpsi(groupId, self._groupsInOpsi):
                        if hasGroupRightParent(groupId, parentGroupId, self._groupsInOpsi):
                            break  # this is the correct group
                        else:
                            groupId = groupId + NAME_ADDITION

                    LOGGER.info("Creating missing group {0!r}".format(groupId))
                    createGroupAction = createHostGroupCommand(groupId, parentGroupId)
                    self._groupsInOpsi.append(getGroupDefinition(createGroupAction))
                    yield createGroupAction

            members = set()
            for member in group.get('uniqueMember', []):
                LOGGER.debug("Member of {!r}: {!r}".format(groupId, member))

                clientId = "{0}.{1}".format(self.getCN(member), getDomain(member))
                clientId = clientId.lower()  # opsi expected FQDN in lowercase.
                LOGGER.debug("Client id is {!r}".format(clientId))
                members.add(clientId)
                if clientId not in self._clientsForProcessing:
                    LOGGER.info("Ignoring unknown member {!r}".format(clientId))
                    continue

                for clientGroupMapping in self._clientGroupMapping:
                    if clientId == clientGroupMapping["objectId"] and groupId == clientGroupMapping["groupId"]:
                        break  # Group assignment is already there
                else:
                    # No existing mapping found so we create one
                    mapping = {
                        "groupId": groupId,
                        "objectId": clientId,
                        "groupType": "HostGroup",
                        "type": "ObjectToGroup"
                    }
                    yield createJSONRPC(
                        "objectToGroup_createObjects",
                        [mapping]
                    )
                    self._clientGroupMapping.append(mapping)

            mappingsToDelete = [
                mapping for mapping in self._clientGroupMapping
                if mapping["groupId"] == groupId and mapping["objectId"] not in members
            ]
            for oldMapping in mappingsToDelete:
                LOGGER.debug("Removing old assignment of {objectId!r} to {groupId!r}".format(**oldMapping))
                yield createJSONRPC(
                    method="objectToGroup_deleteObjects",
                    params=[oldMapping]
                )
                self._clientGroupMapping.remove(oldMapping)

    @staticmethod
    def getCN(dnString):
        cn, _ = dnString.split(',', 1)
        _, cn = cn.split('=', 1)
        return cn
