# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2019 uib GmbH <info@uib.de>

"""
LDAP functionality.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""

import json
from contextlib import contextmanager
from urllib.parse import urlsplit

from ldap3 import Connection, Server

from .logger import LOGGER


__all__ = (
    'checkConnectionToDirectory', 'createSearchQuery', 'convertElementToDict',
    'directoryConnection', 'getDestination', 'getPagedSearchResults',
    'getSearchResults', 'readDirectory'
)

try:
    from ldap3 import SUBTREE
except ImportError:
    SUBTREE = None

DEFAULT_ATTRIBUTES = set(['cn', 'objectCategory', 'objectClass'])


def readDirectory(connection, searchBase, searchQuery, mapping,
    identifyingAttribute, pagedSearch=0):

    LOGGER.debug("LDAP search query: {!r}".format(searchQuery))
    attr = DEFAULT_ATTRIBUTES.union(getAdAttributes(mapping))
    LOGGER.debug("Adding {!r} as identifying attribute".format(identifyingAttribute))
    attr.add(identifyingAttribute)
    LOGGER.debug("Querying for the following attributes: {0}".format(', '.join(attr)))

    for entry in searchDirectory(connection, searchBase, searchQuery, attr, pagedSearch):
        yield entry

    LOGGER.debug("Finished returning elements from directory")


def getAdAttributes(mapping):
    attributeMapping = mapping['client']

    attr = set()
    for opsiAttribute, directoryAttribute in attributeMapping.items():
        if directoryAttribute:
            attr.add(directoryAttribute)
        else:
            LOGGER.warning("No value given for opsi client attribute {!r}".format(opsiAttribute))

    return attr


def createSearchQuery(computerQuery=None):
    """
    Create a search query for our search.

    :raises ValueError: If no queries are given.
    :returntype: str
    """
    if not computerQuery:
        raise ValueError("Missing search configuration!")

    return computerQuery


def searchDirectory(connection, searchBase, query, attributes, pageSize):
    if pageSize:
        searchResults = getPagedSearchResults(connection, searchBase, query, attributes, pageSize=pageSize)
    else:
        searchResults = getSearchResults(connection, searchBase, query, attributes)

    for entry in searchResults:
        LOGGER.debug('yielding element: {!r}'.format(entry))
        yield entry


def getPagedSearchResults(connection, searchBase, searchQuery, attributes, pageSize):
    """
    Return results of the search and use a paged search for searching.
    If a paged search is not possible this falls back to a non-paged search.

    This works as a generator.

    :param connection: The connection to use.
    :type connection: ldap3.Connection
    :param searchBase: Base for the search
    :type searchBase: str
    :param searchQuery: The filter for our search.
    :type searchQuery: str
    :param attributes: The attributes we want.
    :type attributes: [str, ]
    :param pageSize: The number of entries returned per page.
    :type pageSize: int
    """
    if SUBTREE is None:
        LOGGER.warning("Unable to import SUBTREE. Falling back to non-paged-search.")
        entryGenerator = getSearchResults(connection, searchBase, searchQuery, attributes)
    else:
        entryGenerator = connection.extend.standard.paged_search(
            search_base=searchBase,
            search_filter=searchQuery,
            search_scope=SUBTREE,
            attributes=list(attributes),
            paged_size=pageSize,
            generator=True
        )
        LOGGER.debug('Searching for elements in directory with paged search...')

    LOGGER.info('Connected to directory')
    for entry in entryGenerator:
        LOGGER.debug('Found element: {!r}'.format(entry))
        yield entry


def getSearchResults(connection, searchBase, searchQuery, attributes):
    """
    Return results of the search.

    This works as a generator.

    :param connection: The connection to use.
    :type connection: ldap3.Connection
    :param searchBase: Base for the search
    :type searchBase: str
    :param searchQuery: The filter for our search.
    :type searchQuery: str
    :param attributes: The attributes we want.
    :type attributes: [str, ]
    """
    connection.search(searchBase, searchQuery, attributes=list(attributes))
    LOGGER.info('Connected to directory')

    LOGGER.debug('Searching for elements in directory...')
    try:
        for entry in connection.entries:
            LOGGER.debug('Found element: {!r}'.format(entry))
            yield entry
    except AttributeError as attrError:
        LOGGER.debug("Caught attribute error: {0}".format(attrError))
        LOGGER.debug('result: {0}'.format(connection.result))

        LOGGER.debug("Trying fallback for old versions of ldap3...")
        try:
            for entry in connection.response:
                LOGGER.debug('Found element: {!r}'.format(entry))
                yield entry
        except AttributeError as fallBackAttrError:
            LOGGER.debug("Fallback failed: {0!r}".format(fallBackAttrError))

            raise attrError  # We want the original error


def convertElementToDict(element):
    """
    Converts the given `element` to a dict.

    This aims at working with different versions of LDAP3.

    :type element: dict or an element returned from an LDAP search.
    :returntype: dict
    """
    def flattenDict(elementAsDict):
        try:
            # Sometimes you will get an dict with all attributes in
            # another dict inside 'attributes'.
            # We want the content of this directly accessible
            # Since the dict in 'attributes' is a custom
            # case-insensitive dict implementation we only try to use
            # it but we do not check for dict.
            elementAsDict.update(elementAsDict.pop('attributes'))
        except (KeyError, AttributeError):
            pass

        return elementAsDict

    if isinstance(element, dict):
        return flattenDict(element)

    try:
        elemDict = json.loads(element.entry_to_json())
    except AttributeError as attErr:
        LOGGER.debug("Unable to get {0!r} as dict: {1!r}".format(element, attErr))
        LOGGER.debug("The element offers the following: {0}".format(dir(element)))

        try:
            elemDict = element.entry_attributes_as_dict()
        except AttributeError as attrErr:
            LOGGER.debug("Failed to call entry_attributes_as_dict: {!r}".format(attrErr))

            elemDict = element.entry_get_attributes_dict()

    LOGGER.debug("elemDict is: {0!r}".format(elemDict))

    if 'dn' not in elemDict:
        LOGGER.debug("Missing DN. Trying to read...")
        try:
            elemDict['dn'] = element.entry_get_dn()
        except AttributeError as attrErr:
            LOGGER.critical("Unable to get DN for element!")
            raise attrErr

    return flattenDict(elemDict)


def checkConnectionToDirectory(elements):
    index = 0
    try:
        for index, element in enumerate(elements, start=1):
            LOGGER.debug("Picked up element #{0}: {1}".format(index, element))
        LOGGER.info("The current settings matched {0} elements.".format(index))
    except Exception as connectionError:
        LOGGER.exception(connectionError)
        raise connectionError

    if not index:
        message = "No elements found. Does your search criteria fit?"
        LOGGER.warning(message)
        raise RuntimeError(message)


@contextmanager
def directoryConnection(address, user, password, connectionOptions):
    destination = getDestination(address)

    with Connection(destination, user=user, password=password, read_only=True) as connection:
        if connectionOptions.get('start_tls', True):
            connection.start_tls()

        if not connection.bind():
            raise RuntimeError("Connection failed: {0}".format(connection.result))

        yield connection


def getDestination(address):
    destination = address

    if '://' not in destination:
        LOGGER.warning("Directory address misses protocol, adding 'ldap://'.")
        destination = 'ldap://' + destination

    addrInfo = urlsplit(destination)
    LOGGER.debug("Splitted {0!r} to {1!r}".format(destination, addrInfo))
    if addrInfo.port:
        return Server(addrInfo.hostname, port=addrInfo.port)

    return destination
