# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2019 uib GmbH <info@uib.de>

"""
Connection to opsi.

:copyright: uib GmbH <info@uib.de>
:author: Niko Wenselowski <n.wenselowski@uib.de>
"""

import json
import time
from functools import partial

import requests

from . import __version__
from .logger import LOGGER

__all__ = (
    'ModulesError',
    'modulesCheck', 'createJSONRPC', 'checkConnectionToOpsi', 'getSession',
    'getOpsiEndpointAddress', 'getHostFromOpsi', 'updateOpsi', 'checkRootGroup',
    'getIdsOfHostsInOpsi', 'getClientGroupMapping', 'getGroupsFromOpsi',
    'getFromOpsi'
)

USERAGENT = "opsi directory connector/{0}".format(__version__)


class ModulesError(Exception):
    pass


def modulesCheck(session):
    hostCount = len(getFromOpsi(session, "host_getIdents", ['unicode', {"type": "OpsiClient"}]))
    LOGGER.debug("Clients in opsi: {0}".format(hostCount))
    backendInfo = getFromOpsi(session, "backend_info")
    try:
        modules = backendInfo['modules']
        realmodules = backendInfo.get('realmodules', {})
        if not modules:
            raise ModulesError("No Modules detected.")

        if not modules.get('valid'):
            raise ModulesError('Modules file invalid.')

        if not modules.get('customer'):
            raise ModulesError('No customer detected in modules file.')

        if (modules.get('expires', '') != 'never') and (time.mktime(time.strptime(modules.get('expires', '2000-01-01'), "%Y-%m-%d")) - time.time() <= 0):
            raise ModulesError('Modules file expired.')

        if not modules.get('directory-connector'):
            raise ModulesError('Module directory-connector not available in modules.')

        if realmodules and hostCount:
            countInModules = int(realmodules.get('directory-connector', 0))
            if countInModules and hostCount > countInModules + 50:
                raise ModulesError('Client limit {0} reached. Upgrade your license!'.format(countInModules))
    except Exception as verificationError:
        LOGGER.critical("Missing or invalid licensing detected ({0!s}). Please contact sales@uib.de!".format(verificationError))
        raise verificationError


def checkConnectionToOpsi(connectionConfiguration, rootGroup):
    with getSession(**connectionConfiguration) as opsiSession:
        modulesCheck(opsiSession)
        checkRootGroup(opsiSession, rootGroup)
        backendInfo = getFromOpsi(opsiSession, 'backend_info')
        LOGGER.info("Connected to opsi service v{0}".format(backendInfo['opsiVersion']))


def getSession(address, username, password, connectionOptions):
    session = requests.Session()
    session.auth = (username, password)
    session.headers["User-Agent"] = USERAGENT

    opsiEndpoint = getOpsiEndpointAddress(address)

    # We prepare the get method so that the settings for
    # certificate verification and the appropriate address
    # are always given to avoid passing this to any call.
    verifyCertificate = connectionOptions.get('verify_certificate', True)
    LOGGER.debug("Verify certificate? {}".format(verifyCertificate))
    session.get = partial(session.get, opsiEndpoint, verify=verifyCertificate)

    return session


def getOpsiEndpointAddress(address):
    address = address.strip()
    if not address.endswith('/rpc'):
        if not address.endswith('/'):
            address = address + '/'

        address = address + 'rpc'

    return address


def getHostFromOpsi(session, clientId):
    hosts = getFromOpsi(session, 'host_getHashes', [None, {"id": clientId}])
    return hosts[0]


def updateOpsi(session, actions, writeChanges=False, exitOnError=False):
    processableActions = (action for action in actions if action is not None)

    actionCounter = 0
    for actionCounter, action in enumerate(processableActions, start=1):
        LOGGER.info('Processing {!r}'.format(action['method']))

        action['id'] = actionCounter
        LOGGER.debug('Complete call: {!r}'.format(action))
        if writeChanges:
            response = session.get(params=json.dumps(action))
            content = response.json()

            if content['error']:
                errorMessage = "Call to opsi failed: {class} - {message}".format(**content['error'])
                LOGGER.error("ERROR: {}".format(errorMessage))
                if exitOnError:
                    raise RuntimeError(errorMessage)

            LOGGER.debug("Result from call to opsi: {0}".format(content["result"]))
            LOGGER.info('Processing {!r} finished'.format(action['method']))

    LOGGER.debug("Processed {0} action(s) in opsi.".format(actionCounter))


def checkRootGroup(session, name):
    """
    Check if root group exists in opsi.

    :param name: Name of the root group.
    :type name: str
    :raises ValueError: If group does not exist.
    """
    groups = getFromOpsi(session, "group_getHashes", [None, {"id": name}])

    if not groups:
        raise ValueError("Group with id {!r} not found in opsi.".format(name))


def getIdsOfHostsInOpsi(session):
    return set(getFromOpsi(session, "host_getIdents"))


def getClientGroupMapping(session):
    return getFromOpsi(session, "objectToGroup_getHashes")


def getGroupsFromOpsi(session):
    return getFromOpsi(session, "group_getHashes")


def getFromOpsi(session, method, params=None):
    params = json.dumps(createJSONRPC(method, params))
    response = session.get(params=params)
    content = response.json()
    if content['error']:
        raise RuntimeError("Call to opsi failed: {class} - {message}".format(**content['error']))
    LOGGER.debug("get_from_opsi {0}: {1}".format(method, content["result"]))
    return content['result']


def createJSONRPC(method, params=None, identifier=None):
    return {
        "method": method,
        "id": identifier or 1,
        "params": params or []
    }
